from typing_extensions import assert_never

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.hkdf import HKDF

from .crypto_provider import CryptoProvider, HashFunction


__all__ = [
    "CryptoProviderImpl"
]


def get_hash_algorithm(hash_function: HashFunction) -> hashes.HashAlgorithm:
    """
    Args:
        hash_function: Identifier of a hash function.

    Returns:
        The implementation of the hash function as a cryptography
        :class:`~cryptography.hazmat.primitives.hashes.HashAlgorithm` object.
    """

    if hash_function is HashFunction.SHA_256:
        return hashes.SHA256()
    if hash_function is HashFunction.SHA_512:
        return hashes.SHA512()

    return assert_never(hash_function)


class CryptoProviderImpl(CryptoProvider):
    """
    Cryptography provider based on the Python package `cryptography <https://github.com/pyca/cryptography>`_.
    """

    @staticmethod
    async def hkdf_derive(
        hash_function: HashFunction,
        length: int,
        salt: bytes,
        info: bytes,
        key_material: bytes
    ) -> bytes:
        return HKDF(
            algorithm=get_hash_algorithm(hash_function),
            length=length,
            salt=salt,
            info=info,
            backend=default_backend()
        ).derive(key_material)



from __future__ import annotations

from abc import ABC, abstractmethod
import json
import time
import secrets
from typing import FrozenSet, Optional, Set, Tuple, Type, TypeVar, cast

import xeddsa

from .crypto_provider import HashFunction
from .crypto_provider_cryptography import CryptoProviderImpl
from .identity_key_pair import IdentityKeyPair, IdentityKeyPairSeed
from .migrations import parse_base_state_model
from .models import BaseStateModel
from .pre_key_pair import PreKeyPair
from .signed_pre_key_pair import SignedPreKeyPair
from .types import Bundle, IdentityKeyFormat, Header, JSONObject


__all__ = [
    "KeyAgreementException",
    "BaseState"
]


class KeyAgreementException(Exception):
    """
    Exception raised by :meth:`BaseState.get_shared_secret_active` and
    :meth:`BaseState.get_shared_secret_passive` in case of an error related to the key agreement operation.
    """


BaseStateTypeT = TypeVar("BaseStateTypeT", bound="BaseState")


class BaseState(ABC):
    """
    This class is the core of this X3DH implementation. It offers methods to manually manage the X3DH state
    and perform key agreements with other parties.

    Warning:
        This class requires manual state management, including e.g. signed pre key rotation, pre key
        hiding/deletion and refills. The subclass :class:`~x3dh.state.State` automates those
        management/maintenance tasks and should be preferred if external/manual management is not explicitly
        wanted.
    """

    def __init__(self) -> None:
        # Just the type definitions here
        self.__identity_key_format: IdentityKeyFormat
        self.__hash_function: HashFunction
        self.__info: bytes
        self.__identity_key: IdentityKeyPair
        self.__signed_pre_key: SignedPreKeyPair
        self.__old_signed_pre_key: Optional[SignedPreKeyPair]
        self.__pre_keys: Set[PreKeyPair]
        self.__hidden_pre_keys: Set[PreKeyPair]

    @classmethod
    def create(
        cls: Type[BaseStateTypeT],
        identity_key_format: IdentityKeyFormat,
        hash_function: HashFunction,
        info: bytes,
        identity_key_pair: Optional[IdentityKeyPair] = None
    ) -> BaseStateTypeT:
        """
        Args:
            identity_key_format: The format in which the identity public key is included in bundles/headers.
            hash_function: A 256 or 512-bit hash function.
            info: A (byte) string identifying the application.
            identity_key_pair: If set, use the given identity key pair instead of generating a new one.

        Returns:
            A configured instance of :class:`~x3dh.base_state.BaseState`. Note that an identity key pair and a
            signed pre key are generated, but no pre keys. Use :meth:`generate_pre_keys` to generate some.
        """

        self = cls()
        self.__identity_key_format = identity_key_format
        self.__hash_function = hash_function
        self.__info = info
        self.__identity_key = identity_key_pair or IdentityKeyPairSeed(secrets.token_bytes(32))
        self.__signed_pre_key = self.__generate_spk()
        self.__old_signed_pre_key = None
        self.__pre_keys = set()
        self.__hidden_pre_keys = set()

        return self

    ####################
    # abstract methods #
    ####################

    @staticmethod
    @abstractmethod
    def _encode_public_key(key_format: IdentityKeyFormat, pub: bytes) -> bytes:
        """
        Args:
            key_format: The format in which this public key is serialized.
            pub: The public key.

        Returns:
            An encoding of the public key, possibly including information about the curve and type of key,
            though this is application defined. Note that two different public keys must never result in the
            same byte sequence, uniqueness of the public keys must be preserved.
        """

        raise NotImplementedError("Create a subclass of BaseState and implement `_encode_public_key`.")

    #################
    # serialization #
    #################

    @property
    def model(self) -> BaseStateModel:
        """
        Returns:
            The internal state of this :class:`BaseState` as a pydantic model. Note that pre keys hidden using
            :meth:`hide_pre_key` are not considered part of the state.
        """

        return BaseStateModel(
            identity_key=self.__identity_key.model,
            signed_pre_key=self.__signed_pre_key.model,
            old_signed_pre_key=None if self.__old_signed_pre_key is None else self.__old_signed_pre_key.model,
            pre_keys=frozenset(pre_key.priv for pre_key in self.__pre_keys)
        )

    @property
    def json(self) -> JSONObject:
        """
        Returns:
            The internal state of this :class:`BaseState` as a JSON-serializable Python object. Note that pre
            keys hidden using :meth:`hide_pre_key` are not considered part of the state.
        """

        return cast(JSONObject, json.loads(self.model.model_dump_json()))

    @classmethod
    def from_model(
        cls: Type[BaseStateTypeT],
        model: BaseStateModel,
        identity_key_format: IdentityKeyFormat,
        hash_function: HashFunction,
        info: bytes
    ) -> BaseStateTypeT:
        """
        Args:
            model: The pydantic model holding the internal state of a :class:`BaseState`, as produced by
                :attr:`model`.
            identity_key_format: The format in which the identity public key is included in bundles/headers.
            hash_function: A 256 or 512-bit hash function.
            info: A (byte) string identifying the application.

        Returns:
            A configured instance of :class:`BaseState`, with internal state restored from the model.

        Warning:
            Migrations are not provided via the :attr:`model`/:meth:`from_model` API. Use
            :attr:`json`/:meth:`from_json` instead. Refer to :ref:`serialization_and_migration` in the
            documentation for details.
        """

        self = cls()
        self.__identity_key_format = identity_key_format
        self.__hash_function = hash_function
        self.__info = info
        self.__identity_key = IdentityKeyPair.from_model(model.identity_key)
        self.__signed_pre_key = SignedPreKeyPair.from_model(model.signed_pre_key)
        self.__old_signed_pre_key = (
            None
            if model.old_signed_pre_key is None
            else SignedPreKeyPair.from_model(model.old_signed_pre_key)
        )
        self.__pre_keys = { PreKeyPair(pre_key) for pre_key in model.pre_keys }
        self.__hidden_pre_keys = set()

        return self

    @classmethod
    def from_json(
        cls: Type[BaseStateTypeT],
        serialized: JSONObject,
        identity_key_format: IdentityKeyFormat,
        hash_function: HashFunction,
        info: bytes
    ) -> Tuple[BaseStateTypeT, bool]:
        """
        Args:
            serialized: A JSON-serializable Python object holding the internal state of a :class:`BaseState`,
                as produced by :attr:`json`.
            identity_key_format: The format in which the identity public key is included in bundles/headers.
            hash_function: A 256 or 512-bit hash function.
            info: A (byte) string identifying the application.

        Returns:
            A configured instance of :class:`BaseState`, with internal state restored from the serialized
            data, and a flag that indicates whether the bundle needs to be published. The latter was part of
            the pre-stable serialization format.
        """

        model, bundle_needs_publish = parse_base_state_model(serialized)

        self = cls.from_model(
            model,
            identity_key_format,
            hash_function,
            info
        )

        return self, bundle_needs_publish

    #################################
    # key generation and management #
    #################################

    def __generate_spk(self) -> SignedPreKeyPair:
        """
        Returns:
            A newly generated signed pre key.
        """

        # Get the own identity key in the format required for signing, forcing the sign bit if necessary to
        # comply with XEdDSA
        identity_key = self.__identity_key.as_priv().priv
        if self.__identity_key_format is IdentityKeyFormat.CURVE_25519:
            identity_key = xeddsa.priv_force_sign(identity_key, False)

        # Generate the private key of the new signed pre key
        priv = secrets.token_bytes(32)

        # Sign the encoded public key of the new signed pre key
        sig = xeddsa.ed25519_priv_sign(
            identity_key,
            self._encode_public_key(IdentityKeyFormat.CURVE_25519, xeddsa.priv_to_curve25519_pub(priv))
        )

        # Add the current timestamp
        return SignedPreKeyPair(priv=priv, sig=sig, timestamp=int(time.time()))

    @property
    def old_signed_pre_key(self) -> Optional[bytes]:
        """
        Returns:
            The old signed pre key, if there is one.
        """

        return None if self.__old_signed_pre_key is None else self.__old_signed_pre_key.pub

    def signed_pre_key_age(self) -> int:
        """
        Returns:
            The age of the signed pre key, i.e. the time elapsed since it was last rotated, in seconds.
        """

        return int(time.time()) - self.__signed_pre_key.timestamp

    def rotate_signed_pre_key(self) -> None:
        """
        Rotate the signed pre key. Keep the old signed pre key around for one additional rotation period, i.e.
        until this method is called again.
        """

        self.__old_signed_pre_key = self.__signed_pre_key
        self.__signed_pre_key = self.__generate_spk()

    @property
    def hidden_pre_keys(self) -> FrozenSet[bytes]:
        """
        Returns:
            The currently hidden pre keys.
        """

        return frozenset(pre_key.pub for pre_key in self.__hidden_pre_keys)

    def hide_pre_key(self, pre_key_pub: bytes) -> bool:
        """
        Hide a pre key from the bundle returned by :attr:`bundle` and pre key count returned by
        :meth:`get_num_visible_pre_keys`, but keep the pre key for cryptographic operations. Hidden pre keys
        are not included in the serialized state as returned by :attr:`model` and :attr:`json`.

        Args:
            pre_key_pub: The pre key to hide.

        Returns:
            Whether the pre key was visible before and is hidden now.
        """

        hidden_pre_keys = frozenset(filter(lambda pre_key: pre_key.pub == pre_key_pub, self.__pre_keys))

        self.__pre_keys -= hidden_pre_keys
        self.__hidden_pre_keys |= hidden_pre_keys

        return len(hidden_pre_keys) > 0

    def delete_pre_key(self, pre_key_pub: bytes) -> bool:
        """
        Delete a pre key.

        Args:
            pre_key_pub: The pre key to delete. Can be visible or hidden.

        Returns:
            Whether the pre key existed before and is deleted now.
        """

        deleted_pre_keys = frozenset(filter(
            lambda pre_key: pre_key.pub == pre_key_pub,
            self.__pre_keys | self.__hidden_pre_keys
        ))

        self.__pre_keys -= deleted_pre_keys
        self.__hidden_pre_keys -= deleted_pre_keys

        return len(deleted_pre_keys) > 0

    def delete_hidden_pre_keys(self) -> None:
        """
        Delete all pre keys that were previously hidden using :meth:`hide_pre_key`.
        """

        self.__hidden_pre_keys = set()

    def get_num_visible_pre_keys(self) -> int:
        """
        Returns:
            The number of visible pre keys available. The number returned here matches the number of pre keys
            included in the bundle returned by :attr:`bundle`.
        """

        return len(self.__pre_keys)

    def generate_pre_keys(self, num_pre_keys: int) -> None:
        """
        Generate and store pre keys.

        Args:
            num_pre_keys: The number of pre keys to generate.
        """

        for _ in range(num_pre_keys):
            self.__pre_keys.add(PreKeyPair(priv=secrets.token_bytes(32)))

    @property
    def bundle(self) -> Bundle:
        """
        Returns:
            The bundle, i.e. the public information of this state.
        """

        identity_key = self.__identity_key.as_priv().priv

        return Bundle(
            identity_key=(
                xeddsa.priv_to_curve25519_pub(identity_key)
                if self.__identity_key_format is IdentityKeyFormat.CURVE_25519
                else xeddsa.priv_to_ed25519_pub(identity_key)
            ),
            signed_pre_key=self.__signed_pre_key.pub,
            signed_pre_key_sig=self.__signed_pre_key.sig,
            pre_keys=frozenset(pre_key.pub for pre_key in self.__pre_keys)
        )

    #################
    # key agreement #
    #################

    async def get_shared_secret_active(
        self,
        bundle: Bundle,
        associated_data_appendix: bytes = b"",
        require_pre_key: bool = True
    ) -> Tuple[bytes, bytes, Header]:
        """
        Perform an X3DH key agreement, actively.

        Args:
            bundle: The bundle of the passive party.
            associated_data_appendix: Additional information to append to the associated data, like usernames,
                certificates or other identifying information.
            require_pre_key: Use this flag to abort the key agreement if the bundle does not contain a pre
                key.

        Returns:
            The shared secret and associated data shared between both parties, and the header required by the
            other party to complete the passive part of the key agreement.

        Raises:
            KeyAgreementException: If an error occurs during the key agreement. The exception message will
                contain (human-readable) details.
        """

        # Check whether a pre key is required but not included
        if len(bundle.pre_keys) == 0 and require_pre_key:
            raise KeyAgreementException("This bundle does not contain a pre key.")

        # Get the identity key of the other party in the format required for signature verification
        other_identity_key = bundle.identity_key
        if self.__identity_key_format is IdentityKeyFormat.CURVE_25519:
            other_identity_key = xeddsa.curve25519_pub_to_ed25519_pub(other_identity_key, False)

        # Verify the signature on the signed pre key of the other party
        if not xeddsa.ed25519_verify(
            bundle.signed_pre_key_sig,
            other_identity_key,
            self._encode_public_key(IdentityKeyFormat.CURVE_25519, bundle.signed_pre_key)
        ):
            raise KeyAgreementException("The signature of the signed pre key could not be verified.")

        # All pre-checks successful.

        # Choose a pre key if available
        pre_key = None if len(bundle.pre_keys) == 0 else secrets.choice(list(bundle.pre_keys))

        # Generate the ephemeral key required for the key agreement
        ephemeral_key = secrets.token_bytes(32)

        # Get the own identity key in the format required for X25519
        own_identity_key = self.__identity_key.as_priv().priv

        # Get the identity key of the other party in the format required for X25519
        other_identity_key = bundle.identity_key
        if self.__identity_key_format is IdentityKeyFormat.ED_25519:
            other_identity_key = xeddsa.ed25519_pub_to_curve25519_pub(other_identity_key)

        # Calculate the three to four Diffie-Hellman shared secrets that become the input of HKDF in the next
        # step
        dh1 = xeddsa.x25519(own_identity_key, bundle.signed_pre_key)
        dh2 = xeddsa.x25519(ephemeral_key, other_identity_key)
        dh3 = xeddsa.x25519(ephemeral_key, bundle.signed_pre_key)
        dh4 = b"" if pre_key is None else xeddsa.x25519(ephemeral_key, pre_key)

        # Prepare salt and padding
        salt = b"\x00" * self.__hash_function.hash_size
        padding = b"\xFF" * 32

        # Use HKDF to derive the final shared secret
        shared_secret = await CryptoProviderImpl.hkdf_derive(
            self.__hash_function,
            32,
            salt,
            self.__info,
            padding + dh1 + dh2 + dh3 + dh4
        )

        # Build the associated data for further use by other protocols
        associated_data = (
            self._encode_public_key(self.__identity_key_format, self.bundle.identity_key)
            + self._encode_public_key(self.__identity_key_format, bundle.identity_key)
            + associated_data_appendix
        )

        # Build the header required by the other party to complete the passive part of the key agreement
        header = Header(
            identity_key=self.bundle.identity_key,
            ephemeral_key=xeddsa.priv_to_curve25519_pub(ephemeral_key),
            pre_key=pre_key,
            signed_pre_key=bundle.signed_pre_key
        )

        return shared_secret, associated_data, header

    async def get_shared_secret_passive(
        self,
        header: Header,
        associated_data_appendix: bytes = b"",
        require_pre_key: bool = True
    ) -> Tuple[bytes, bytes, SignedPreKeyPair]:
        """
        Perform an X3DH key agreement, passively.

        Args:
            header: The header received from the active party.
            associated_data_appendix: Additional information to append to the associated data, like usernames,
                certificates or other identifying information.
            require_pre_key: Use this flag to abort the key agreement if the active party did not use a pre
                key.

        Returns:
            The shared secret and the associated data shared between both parties, and the signed pre key pair
            that was used during the key exchange, for use by follow-up protocols.

        Raises:
            KeyAgreementException: If an error occurs during the key agreement. The exception message will
                contain (human-readable) details.
        """

        # Check whether the signed pre key used by this initiation is still available
        signed_pre_key: Optional[SignedPreKeyPair] = None

        if header.signed_pre_key == self.__signed_pre_key.pub:
            # The current signed pre key was used
            signed_pre_key = self.__signed_pre_key

        if self.__old_signed_pre_key is not None and header.signed_pre_key == self.__old_signed_pre_key.pub:
            # The old signed pre key was used
            signed_pre_key = self.__old_signed_pre_key

        if signed_pre_key is None:
            raise KeyAgreementException(
                "This key agreement attempt uses a signed pre key that is not available any more."
            )

        # Check whether a pre key is required but not used
        if header.pre_key is None and require_pre_key:
            raise KeyAgreementException("This key agreement attempt does not use a pre key.")

        # If a pre key was used, check whether it is still available
        pre_key: Optional[bytes] = None
        if header.pre_key is not None:
            pre_key = next((
                pre_key.priv
                for pre_key
                in self.__pre_keys | self.__hidden_pre_keys
                if pre_key.pub == header.pre_key
            ), None)

            if pre_key is None:
                raise KeyAgreementException(
                    "This key agreement attempt uses a pre key that is not available any more."
                )

        # Get the own identity key in the format required for X25519
        own_identity_key = self.__identity_key.as_priv().priv

        # Get the identity key of the other party in the format required for X25519
        other_identity_key = header.identity_key
        if self.__identity_key_format is IdentityKeyFormat.ED_25519:
            other_identity_key = xeddsa.ed25519_pub_to_curve25519_pub(other_identity_key)

        # Calculate the three to four Diffie-Hellman shared secrets that become the input of HKDF in the next
        # step
        dh1 = xeddsa.x25519(signed_pre_key.priv, other_identity_key)
        dh2 = xeddsa.x25519(own_identity_key, header.ephemeral_key)
        dh3 = xeddsa.x25519(signed_pre_key.priv, header.ephemeral_key)
        dh4 = b"" if pre_key is None else xeddsa.x25519(pre_key, header.ephemeral_key)

        # Prepare salt and padding
        salt = b"\x00" * self.__hash_function.hash_size
        padding = b"\xFF" * 32

        # Use HKDF to derive the final shared secret
        shared_secret = await CryptoProviderImpl.hkdf_derive(
            self.__hash_function,
            32,
            salt,
            self.__info,
            padding + dh1 + dh2 + dh3 + dh4
        )

        # Build the associated data for further use by other protocols
        associated_data = (
            self._encode_public_key(self.__identity_key_format, header.identity_key)
            + self._encode_public_key(self.__identity_key_format, self.bundle.identity_key)
            + associated_data_appendix
        )

        return shared_secret, associated_data, signed_pre_key



from __future__ import annotations

from abc import abstractmethod
from typing import Optional, Tuple, Type, TypeVar

from .base_state import BaseState
from .crypto_provider import HashFunction
from .identity_key_pair import IdentityKeyPair
from .migrations import parse_base_state_model
from .models import BaseStateModel
from .signed_pre_key_pair import SignedPreKeyPair
from .types import Bundle, IdentityKeyFormat, Header, JSONObject


__all__ = [
    "State"
]


StateTypeT = TypeVar("StateTypeT", bound="State")


class State(BaseState):
    """
    This class is the core of this X3DH implementation. It manages the own :class:`~x3dh.types.Bundle` and
    offers methods to perform key agreements with other parties. Use :class:`~x3dh.base_state.BaseState`
    directly if manual state management is needed. Note that you can still use the methods available for
    manual state management, but doing so shouldn't be required.

    Warning:
        :meth:`rotate_signed_pre_key` should be called periodically to check whether the signed pre key needs
        to be rotated and to perform the rotation if necessary.
    """

    def __init__(self) -> None:
        super().__init__()

        # Just the type definitions here
        self.__signed_pre_key_rotation_period: int
        self.__pre_key_refill_threshold: int
        self.__pre_key_refill_target: int

    @classmethod
    def create(
        cls: Type[StateTypeT],
        identity_key_format: IdentityKeyFormat,
        hash_function: HashFunction,
        info: bytes,
        identity_key_pair: Optional[IdentityKeyPair] = None,
        signed_pre_key_rotation_period: int = 7 * 24 * 60 * 60,
        pre_key_refill_threshold: int = 99,
        pre_key_refill_target: int = 100
    ) -> StateTypeT:
        """
        Args:
            identity_key_format: The format in which the identity public key is included in bundles/headers.
            hash_function: A 256 or 512-bit hash function.
            info: A (byte) string identifying the application.
            signed_pre_key_rotation_period: Rotate the signed pre key after this amount of time in seconds.
            pre_key_refill_threshold: Threshold for refilling the pre keys.
            pre_key_refill_target: When less then ``pre_key_refill_threshold`` pre keys are available,
                generate new ones until there are ``pre_key_refill_target`` pre keys again.
            identity_key_pair: If set, use the given identity key pair instead of generating a new one.

        Returns:
            A configured instance of :class:`~x3dh.state.State`.
        """
        # pylint: disable=protected-access

        if signed_pre_key_rotation_period < 1:
            raise ValueError(
                "Invalid value passed for the `signed_pre_key_rotation_period` parameter. The signed pre key"
                " rotation period must be at least one day."
            )

        if not 1 <= pre_key_refill_threshold <= pre_key_refill_target:
            raise ValueError(
                "Invalid value(s) passed for the `pre_key_refill_threshold` / `pre_key_refill_target`"
                " parameter(s). `pre_key_refill_threshold` must be greater than or equal to '1' and lower"
                " than or equal to `pre_key_refill_target`."
            )

        self = super().create(identity_key_format, hash_function, info, identity_key_pair)

        self.__signed_pre_key_rotation_period = signed_pre_key_rotation_period
        self.__pre_key_refill_threshold = pre_key_refill_threshold
        self.__pre_key_refill_target = pre_key_refill_target

        self.generate_pre_keys(pre_key_refill_target)

        # I believe this is a false positive by pylint
        self._publish_bundle(self.bundle)  # pylint: disable=no-member

        return self

    ####################
    # abstract methods #
    ####################

    @abstractmethod
    def _publish_bundle(self, bundle: Bundle) -> None:
        """
        Args:
            bundle: The bundle to publish, overwriting previously published data.

        Note:
            In addition to publishing the bundle, this method can be used as a trigger to persist the state.
            Persisting the state in this method guarantees always remaining up-to-date.

        Note:
            This method is called from :meth:`create`, before :meth:`create` has returned the instance. Thus,
            modifications to the object (``self``, in case of subclasses) may not have happened when this
            method is called.

        Note:
            Even though this method is expected to perform I/O, it is deliberately not marked as async, since
            completion of the I/O operation is not a requirement for the program flow to continue, and making
            this method async would complicate API design with regards to inheritance from
            :class:`~x3dh.base_state.BaseState`.
        """

        raise NotImplementedError("Create a subclass of State and implement `_publish_bundle`.")

    #################
    # serialization #
    #################

    @classmethod
    def from_model(
        cls: Type[StateTypeT],
        model: BaseStateModel,
        identity_key_format: IdentityKeyFormat,
        hash_function: HashFunction,
        info: bytes,
        signed_pre_key_rotation_period: int = 7 * 24 * 60 * 60,
        pre_key_refill_threshold: int = 99,
        pre_key_refill_target: int = 100
    ) -> StateTypeT:
        """
        Args:
            model: The pydantic model holding the internal state of a :class:`State`, as produced by
                :attr:`~x3dh.base_state.BaseState.model`.
            identity_key_format: The format in which the identity public key is included in bundles/headers.
            hash_function: A 256 or 512-bit hash function.
            info: A (byte) string identifying the application.
            signed_pre_key_rotation_period: Rotate the signed pre key after this amount of time in seconds.
            pre_key_refill_threshold: Threshold for refilling the pre keys.
            pre_key_refill_target: When less then ``pre_key_refill_threshold`` pre keys are available,
                generate new ones until there are ``pre_key_refill_target`` pre keys again.

        Returns:
            A configured instance of :class:`State`, with internal state restored from the model.

        Warning:
            Migrations are not provided via the :attr:`~x3dh.base_state.BaseState.model`/:meth:`from_model`
            API. Use :attr:`~x3dh.base_state.BaseState.json`/:meth:`from_json` instead. Refer to
            :ref:`serialization_and_migration` in the documentation for details.
        """
        # pylint: disable=protected-access

        if signed_pre_key_rotation_period < 1:
            raise ValueError(
                "Invalid value passed for the `signed_pre_key_rotation_period` parameter. The signed pre key"
                " rotation period must be at least one day."
            )

        if not 1 <= pre_key_refill_threshold <= pre_key_refill_target:
            raise ValueError(
                "Invalid value(s) passed for the `pre_key_refill_threshold` / `pre_key_refill_target`"
                " parameter(s). `pre_key_refill_threshold` must be greater than or equal to '1' and lower"
                " than or equal to `pre_key_refill_target`."
            )

        self = super().from_model(model, identity_key_format, hash_function, info)

        self.__signed_pre_key_rotation_period = signed_pre_key_rotation_period
        self.__pre_key_refill_threshold = pre_key_refill_threshold
        self.__pre_key_refill_target = pre_key_refill_target

        self.rotate_signed_pre_key()

        return self

    @classmethod
    def from_json(
        cls: Type[StateTypeT],
        serialized: JSONObject,
        identity_key_format: IdentityKeyFormat,
        hash_function: HashFunction,
        info: bytes,
        signed_pre_key_rotation_period: int = 7 * 24 * 60 * 60,
        pre_key_refill_threshold: int = 99,
        pre_key_refill_target: int = 100
    ) -> Tuple[StateTypeT, bool]:
        """
        Args:
            serialized: A JSON-serializable Python object holding the internal state of a :class:`State`,
                as produced by :attr:`~x3dh.base_state.BaseState.json`.
            identity_key_format: The format in which the identity public key is included in bundles/headers.
            hash_function: A 256 or 512-bit hash function.
            info: A (byte) string identifying the application.
            signed_pre_key_rotation_period: Rotate the signed pre key after this amount of time in seconds.
            pre_key_refill_threshold: Threshold for refilling the pre keys.
            pre_key_refill_target: When less then ``pre_key_refill_threshold`` pre keys are available,
                generate new ones until there are ``pre_key_refill_target`` pre keys again.

        Returns:
            A configured instance of :class:`State`, with internal state restored from the serialized data,
            and a flag that indicates whether the bundle needed to be published. The latter was part of the
            pre-stable serialization format and is handled automatically by this :meth:`from_json`
            implementation.
        """
        # pylint: disable=protected-access

        model, bundle_needs_publish = parse_base_state_model(serialized)

        self = cls.from_model(
            model,
            identity_key_format,
            hash_function,
            info,
            signed_pre_key_rotation_period,
            pre_key_refill_threshold,
            pre_key_refill_target
        )

        if bundle_needs_publish:
            # I believe this is a false positive by pylint
            self._publish_bundle(self.bundle)  # pylint: disable=no-member

        return self, False

    #################################
    # key generation and management #
    #################################

    def rotate_signed_pre_key(self, force: bool = False) -> None:
        """
        Check whether the signed pre key is due for rotation, and rotate it if necessary. Call this method
        periodically to make sure the signed pre key is always up to date.

        Args:
            force: Whether to force rotation regardless of the age of the current signed pre key.
        """

        if force or self.signed_pre_key_age() > self.__signed_pre_key_rotation_period:
            super().rotate_signed_pre_key()

            self._publish_bundle(self.bundle)

    #################
    # key agreement #
    #################

    async def get_shared_secret_passive(
        self,
        header: Header,
        associated_data_appendix: bytes = b"",
        require_pre_key: bool = True
    ) -> Tuple[bytes, bytes, SignedPreKeyPair]:
        """
        Perform an X3DH key agreement, passively.

        Args:
            header: The header received from the active party.
            associated_data_appendix: Additional information to append to the associated data, like usernames,
                certificates or other identifying information.
            require_pre_key: Use this flag to abort the key agreement if the active party did not use a pre
                key.

        Returns:
            The shared secret and the associated data shared between both parties, and the signed pre key pair
            that was used during the key exchange, for use by follow-up protocols.

        Raises:
            KeyAgreementException: If an error occurs during the key agreement. The exception message will
                contain (human-readable) details.
        """

        shared_secret, associated_data, signed_pre_key_pair = await super().get_shared_secret_passive(
            header,
            associated_data_appendix,
            require_pre_key
        )

        # If a pre key was used, remove it from the pool and refill the pool if necessary
        if header.pre_key is not None:
            self.delete_pre_key(header.pre_key)

            if self.get_num_visible_pre_keys() < self.__pre_key_refill_threshold:
                self.generate_pre_keys(self.__pre_key_refill_target - self.get_num_visible_pre_keys())

            self._publish_bundle(self.bundle)

        return shared_secret, associated_data, signed_pre_key_pair



# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a full list see
# the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory, add these
# directories to sys.path here. If the directory is relative to the documentation root,
# use os.path.abspath to make it absolute, like shown here.
import os
import sys

root_dir_path = os.path.abspath(os.path.join(os.path.dirname(os.path.abspath(__file__)), ".."))
sys.path.append(os.path.join(root_dir_path, "x3dh"))

# -- Project information -----------------------------------------------------------------

from version import __version__ as __version

import tomllib
from typing import List

with open(os.path.join(root_dir_path, "pyproject.toml"), "rb") as f:
    __pyproject = tomllib.load(f)

project: str   = __pyproject["project"]["name"]
author: str    = __pyproject["project"]["authors"][0]["name"]
copyright: str = f"2026, {author}"

__classifiers: List[str] = __pyproject["project"]["classifiers"]

__tag: str = "unknown"
if "Development Status :: 3 - Alpha" in __classifiers:
    __tag = "alpha"
if "Development Status :: 4 - Beta" in __classifiers:
    __tag = "beta"
if "Development Status :: 5 - Production/Stable" in __classifiers:
    __tag = "stable"

# The short X.Y version
version = __version
# The full version, including alpha/beta/rc tags
release = f"{__version}-{__tag}"

# -- General configuration ---------------------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be extensions coming
# with Sphinx (named "sphinx.ext.*") or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx_autodoc_typehints"
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "pydantic": ("https://docs.pydantic.dev/latest", None)
}

# Add any paths that contain templates here, relative to this directory.
templates_path = [ "_templates" ]

# List of patterns, relative to source directory, that match files and directories to
# ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [ "_build", "Thumbs.db", ".DS_Store" ]

# -- Options for HTML output -------------------------------------------------------------

# The theme to use for HTML and HTML Help pages. See the documentation for a list of
# builtin themes.
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here, relative to
# this directory. They are copied after the builtin static files, so a file named
# "default.css" will overwrite the builtin "default.css".
html_static_path = [ "_static" ]

# -- Autodoc Configuration ---------------------------------------------------------------

autodoc_typehints = "description"
autodoc_type_aliases = { k: k for k in {
    "JSONType",
    "JSONObject"
} }

# https://github.com/sphinx-doc/sphinx/issues/10785
def resolve_type_aliases(app, env, node, contnode):
    """Resolve :class: references to our type aliases as :attr: instead."""
    if (
        node["refdomain"] == "py"
        and node["reftype"] == "class"
        and node["reftarget"] in autodoc_type_aliases
    ):
        return app.env.get_domain("py").resolve_xref(
            env, node["refdoc"], app.builder, "attr", node["reftarget"], node, contnode
        )

def autodoc_skip_member_handler(app, what, name, obj, skip, options):
    # Skip private members, i.e. those that start with double underscores but do not end in underscores
    if name.startswith("__") and not name.endswith("_"):
        return True

    # Other fixed names to always skip
    if name in { "_abc_impl", "model_config" }:
        return True

    # Skip __init__s without documentation. Those are just used for type hints.
    if name == "__init__" and obj.__doc__ is None:
        return True

    return None

def setup(app):
    app.connect("autodoc-skip-member", autodoc_skip_member_handler)
    app.connect("missing-reference", resolve_type_aliases)



from __future__ import annotations

from abc import ABC, abstractmethod
import json
from typing import cast
from typing_extensions import assert_never

import xeddsa

from .migrations import parse_identity_key_pair_model
from .models import IdentityKeyPairModel
from .types import JSONObject, SecretType


__all__ = [
    "IdentityKeyPair",
    "IdentityKeyPairPriv",
    "IdentityKeyPairSeed"
]


class IdentityKeyPair(ABC):
    """
    An identity key pair.

    There are following requirements for the identity key pair:

    * It must be able to create and verify Ed25519-compatible signatures.
    * It must be able to perform X25519-compatible Diffie-Hellman key agreements.

    There are at least two different kinds of key pairs that can fulfill these requirements: Ed25519 key pairs
    and Curve25519 key pairs. The birational equivalence of both curves can be used to "convert" one pair to
    the other.

    Both types of key pairs share the same private key, however instead of a private key, a seed can be used
    which the private key is derived from using SHA-512. This is standard practice for Ed25519, where the
    other 32 bytes of the SHA-512 seed hash are used as a nonce during signing. If a new key pair has to be
    generated, this implementation generates a seed.
    """

    @property
    def model(self) -> IdentityKeyPairModel:
        """
        Returns:
            The internal state of this :class:`IdentityKeyPair` as a pydantic model.
        """

        return IdentityKeyPairModel(secret=self.secret, secret_type=self.secret_type)

    @property
    def json(self) -> JSONObject:
        """
        Returns:
            The internal state of this :class:`IdentityKeyPair` as a JSON-serializable Python object.
        """

        return cast(JSONObject, json.loads(self.model.json()))

    @staticmethod
    def from_model(model: IdentityKeyPairModel) -> "IdentityKeyPair":
        """
        Args:
            model: The pydantic model holding the internal state of an :class:`IdentityKeyPair`, as produced
                by :attr:`model`.

        Returns:
            A configured instance of :class:`IdentityKeyPair`, with internal state restored from the model.

        Warning:
            Migrations are not provided via the :attr:`model`/:meth:`from_model` API. Use
            :attr:`json`/:meth:`from_json` instead. Refer to :ref:`serialization_and_migration` in the
            documentation for details.
        """

        if model.secret_type is SecretType.PRIV:
            return IdentityKeyPairPriv(model.secret)
        if model.secret_type is SecretType.SEED:
            return IdentityKeyPairSeed(model.secret)

        return assert_never(model.secret_type)

    @staticmethod
    def from_json(serialized: JSONObject) -> "IdentityKeyPair":
        """
        Args:
            serialized: A JSON-serializable Python object holding the internal state of an
                :class:`IdentityKeyPair`, as produced by :attr:`json`.

        Returns:
            A configured instance of :class:`IdentityKeyPair`, with internal state restored from the
            serialized data.
        """

        return IdentityKeyPair.from_model(parse_identity_key_pair_model(serialized))

    @property
    @abstractmethod
    def secret_type(self) -> SecretType:
        """
        Returns:
            The type of secret used by this identity key (i.e. a seed or private key).
        """

    @property
    @abstractmethod
    def secret(self) -> bytes:
        """
        Returns:
            The secret used by this identity key, i.e. the seed or private key.
        """

    @abstractmethod
    def as_priv(self) -> "IdentityKeyPairPriv":
        """
        Returns:
            An :class:`IdentityKeyPairPriv` derived from this instance, or the instance itself if it already
            is an :class:`IdentityKeyPairPriv`.
        """


class IdentityKeyPairPriv(IdentityKeyPair):
    """
    An :class:`IdentityKeyPair` represented by a Curve25519/Ed25519 private key.
    """

    def __init__(self, priv: bytes) -> None:
        """
        Args:
            priv: The Curve25519/Ed25519 private key.
        """

        if len(priv) != 32:
            raise ValueError("Expected the private key to be 32 bytes long.")

        self.__priv = priv

    @property
    def secret_type(self) -> SecretType:
        return SecretType.PRIV

    @property
    def secret(self) -> bytes:
        return self.priv

    def as_priv(self) -> "IdentityKeyPairPriv":
        return self

    @property
    def priv(self) -> bytes:
        """
        Returns:
            The Curve25519/Ed25519 private key.
        """

        return self.__priv


class IdentityKeyPairSeed(IdentityKeyPair):
    """
    An :class:`IdentityKeyPair` represented by a Curve25519/Ed25519 seed.
    """

    def __init__(self, seed: bytes) -> None:
        """
        Args:
            seed: The Curve25519/Ed25519 seed.
        """

        if len(seed) != 32:
            raise ValueError("Expected the seed to be 32 bytes long.")

        self.__seed = seed

    @property
    def secret_type(self) -> SecretType:
        return SecretType.SEED

    @property
    def secret(self) -> bytes:
        return self.seed

    def as_priv(self) -> "IdentityKeyPairPriv":
        return IdentityKeyPairPriv(xeddsa.seed_to_priv(self.__seed))

    @property
    def seed(self) -> bytes:
        """
        Returns:
            The Curve25519/Ed25519 seed.
        """

        return self.__seed



from .version import __version__ as __version__

from .base_state import (
    KeyAgreementException as KeyAgreementException,
    BaseState as BaseState
)
from .crypto_provider import HashFunction as HashFunction
from .models import (
    BaseStateModel as BaseStateModel,
    IdentityKeyPairModel as IdentityKeyPairModel,
    SignedPreKeyPairModel as SignedPreKeyPairModel
)
from .state import State as State
from .types import (
    Bundle as Bundle,
    Header as Header,
    IdentityKeyFormat as IdentityKeyFormat,
    JSONType as JSONType,
    JSONObject as JSONObject
)



from typing import Any, FrozenSet, Optional

from pydantic import BaseModel
from pydantic.functional_serializers import PlainSerializer
from pydantic.functional_validators import PlainValidator
from typing_extensions import Annotated

from .types import SecretType


__all__ = [
    "BaseStateModel",
    "IdentityKeyPairModel",
    "SignedPreKeyPairModel"
]


def _json_bytes_decoder(val: Any) -> bytes:
    """
    Decode bytes from a string according to the JSON specification. See
    https://github.com/samuelcolvin/pydantic/issues/3756 for details.

    Args:
        val: The value to type check and decode.

    Returns:
        The value decoded to bytes. If the value is bytes already, it is returned unmodified.

    Raises:
        ValueError: if the value is not correctly encoded.
    """

    if isinstance(val, bytes):
        return val
    if isinstance(val, str):
        return bytes(map(ord, val))
    raise ValueError("bytes fields must be encoded as bytes or str.")


def _json_bytes_encoder(val: bytes) -> str:
    """
    Encode bytes as a string according to the JSON specification. See
    https://github.com/samuelcolvin/pydantic/issues/3756 for details.

    Args:
        val: The bytes to encode.

    Returns:
        The encoded bytes.
    """

    return "".join(map(chr, val))


JsonBytes = Annotated[bytes, PlainValidator(_json_bytes_decoder), PlainSerializer(_json_bytes_encoder)]


class IdentityKeyPairModel(BaseModel):
    """
    The model representing the internal state of an :class:`~x3dh.identity_key_pair.IdentityKeyPair`.
    """

    version: str = "1.0.0"
    secret: JsonBytes
    secret_type: SecretType


class SignedPreKeyPairModel(BaseModel):
    """
    The model representing the internal state of a :class:`~x3dh.signed_pre_key_pair.SignedPreKeyPair`.
    """

    version: str = "1.0.0"
    priv: JsonBytes
    sig: JsonBytes
    timestamp: int


class BaseStateModel(BaseModel):
    """
    The model representing the internal state of a :class:`~x3dh.base_state.BaseState`.
    """

    version: str = "1.0.0"
    identity_key: IdentityKeyPairModel
    signed_pre_key: SignedPreKeyPairModel
    old_signed_pre_key: Optional[SignedPreKeyPairModel]
    pre_keys: FrozenSet[JsonBytes]



from abc import ABC, abstractmethod
import enum
from typing_extensions import assert_never


__all__ = [
    "CryptoProvider",
    "HashFunction"
]


@enum.unique
class HashFunction(enum.Enum):
    """
    Enumeration of the hash functions supported for the key derivation step of X3DH.
    """

    SHA_256 = "SHA_256"
    SHA_512 = "SHA_512"

    @property
    def hash_size(self) -> int:
        """
        Returns:
            The byte size of the hashes produced by this hash function.
        """

        if self is HashFunction.SHA_256:
            return 32
        if self is HashFunction.SHA_512:
            return 64

        return assert_never(self)


class CryptoProvider(ABC):
    """
    Abstraction of the cryptographic operations needed by this package to allow for different backend
    implementations.
    """

    @staticmethod
    @abstractmethod
    async def hkdf_derive(
        hash_function: HashFunction,
        length: int,
        salt: bytes,
        info: bytes,
        key_material: bytes
    ) -> bytes:
        """
        Args:
            hash_function: The hash function to parameterize the HKDF with.
            length: The number of bytes to derive.
            salt: The salt input for the HKDF.
            info: The info input for the HKDF.
            key_material: The input key material to derive from.

        Returns:
            The derived key material.
        """



from __future__ import annotations

import json
from typing import NamedTuple, cast

import xeddsa

from .migrations import parse_signed_pre_key_pair_model
from .models import SignedPreKeyPairModel
from .types import JSONObject


__all__ = [
    "SignedPreKeyPair"
]


class SignedPreKeyPair(NamedTuple):
    """
    A signed pre key, i.e. a pre key whose public key was encoded using an application-specific encoding
    format, then signed by the identity key, and stored together with a generation timestamp for periodic
    rotation.
    """

    priv: bytes
    sig: bytes
    timestamp: int

    @property
    def pub(self) -> bytes:
        """
        Returns:
            The public key of this signed pre key.
        """

        return xeddsa.priv_to_curve25519_pub(self.priv)

    @property
    def model(self) -> SignedPreKeyPairModel:
        """
        Returns:
            The internal state of this :class:`SignedPreKeyPair` as a pydantic model.
        """

        return SignedPreKeyPairModel(priv=self.priv, sig=self.sig, timestamp=self.timestamp)

    @property
    def json(self) -> JSONObject:
        """
        Returns:
            The internal state of this :class:`SignedPreKeyPair` as a JSON-serializable Python object.
        """

        return cast(JSONObject, json.loads(self.model.json()))

    @staticmethod
    def from_model(model: SignedPreKeyPairModel) -> "SignedPreKeyPair":
        """
        Args:
            model: The pydantic model holding the internal state of a :class:`SignedPreKeyPair`, as produced
                by :attr:`model`.

        Returns:
            A configured instance of :class:`SignedPreKeyPair`, with internal state restored from the model.

        Warning:
            Migrations are not provided via the :attr:`model`/:meth:`from_model` API. Use
            :attr:`json`/:meth:`from_json` instead. Refer to :ref:`serialization_and_migration` in the
            documentation for details.
        """

        return SignedPreKeyPair(priv=model.priv, sig=model.sig, timestamp=model.timestamp)

    @staticmethod
    def from_json(serialized: JSONObject) -> "SignedPreKeyPair":
        """
        Args:
            serialized: A JSON-serializable Python object holding the internal state of a
                :class:`SignedPreKeyPair`, as produced by :attr:`json`.

        Returns:
            A configured instance of :class:`SignedPreKeyPair`, with internal state restored from the
            serialized data.
        """

        return SignedPreKeyPair.from_model(parse_signed_pre_key_pair_model(serialized))



from typing import NamedTuple

import xeddsa


__all__ = [
    "PreKeyPair"
]


class PreKeyPair(NamedTuple):
    """
    A pre key.
    """

    priv: bytes

    @property
    def pub(self) -> bytes:
        """
        Returns:
            The public key of this pre key.
        """

        return xeddsa.priv_to_curve25519_pub(self.priv)



from __future__ import annotations

import base64
from typing import List, Tuple, cast

from pydantic import BaseModel

from .models import IdentityKeyPairModel, SignedPreKeyPairModel, BaseStateModel
from .types import JSONObject, SecretType


__all__ = [
    "parse_identity_key_pair_model",
    "parse_signed_pre_key_pair_model",
    "parse_base_state_model"
]


class PreStableKeyPairModel(BaseModel):
    """
    This model describes how a key pair was serialized in pre-stable serialization format.
    """

    priv: str
    pub: str


class PreStableSignedPreKeyModel(BaseModel):
    """
    This model describes how a signed pre-key was serialized in pre-stable serialization format.
    """

    key: PreStableKeyPairModel
    signature: str
    timestamp: float


class PreStableModel(BaseModel):
    """
    This model describes how State instances were serialized in pre-stable serialization format.
    """

    changed: bool
    ik: PreStableKeyPairModel
    spk: PreStableSignedPreKeyModel
    otpks: List[PreStableKeyPairModel]


def parse_identity_key_pair_model(serialized: JSONObject) -> IdentityKeyPairModel:
    """
    Parse a serialized :class:`~x3dh.identity_key_pair.IdentityKeyPair` instance, as returned by
    :attr:`~x3dh.identity_key_pair.IdentityKeyPair.json`, into the most recent pydantic model available for
    the class. Perform migrations in case the pydantic models were updated.

    Args:
        serialized: The serialized instance.

    Returns:
        The model, which can be used to restore the instance using
        :meth:`~x3dh.identity_key_pair.IdentityKeyPair.from_model`.

    Note:
        Pre-stable data can only be migrated as a whole using :func:`parse_base_state_model`.
    """

    # Each model has a Python string "version" in its root. Use that to find the model that the data was
    # serialized from.
    version = cast(str, serialized["version"])
    model: BaseModel = {
        "1.0.0": IdentityKeyPairModel,
        "1.0.1": IdentityKeyPairModel
    }[version](**serialized)  # type: ignore[arg-type]

    # Once all migrations have been applied, the model should be an instance of the most recent model
    assert isinstance(model, IdentityKeyPairModel)

    return model


def parse_signed_pre_key_pair_model(serialized: JSONObject) -> SignedPreKeyPairModel:
    """
    Parse a serialized :class:`~x3dh.signed_pre_key_pair.SignedPreKeyPair` instance, as returned by
    :attr:`~x3dh.signed_pre_key_pair.SignedPreKeyPair.json`, into the most recent pydantic model available for
    the class. Perform migrations in case the pydantic models were updated.

    Args:
        serialized: The serialized instance.

    Returns:
        The model, which can be used to restore the instance using
        :meth:`~x3dh.signed_pre_key_pair.SignedPreKeyPair.from_model`.

    Note:
        Pre-stable data can only be migrated as a whole using :func:`parse_base_state_model`.
    """

    # Each model has a Python string "version" in its root. Use that to find the model that the data was
    # serialized from.
    version = cast(str, serialized["version"])
    model: BaseModel = {
        "1.0.0": SignedPreKeyPairModel,
        "1.0.1": SignedPreKeyPairModel
    }[version](**serialized)  # type: ignore[arg-type]

    # Once all migrations have been applied, the model should be an instance of the most recent model
    assert isinstance(model, SignedPreKeyPairModel)

    return model


def parse_base_state_model(serialized: JSONObject) -> Tuple[BaseStateModel, bool]:
    """
    Parse a serialized :class:`~x3dh.base_state.BaseState` instance, as returned by
    :attr:`~x3dh.base_state.BaseState.json`, into the most recent pydantic model available for the class.
    Perform migrations in case the pydantic models were updated. Supports migration of pre-stable data.

    Args:
        serialized: The serialized instance.

    Returns:
        The model, which can be used to restore the instance using
        :meth:`~x3dh.base_state.BaseState.from_model`, and a flag that indicates whether the bundle needs to
        be published, which was part of the pre-stable serialization format.
    """

    bundle_needs_publish = False

    # Each model has a Python string "version" in its root. Use that to find the model that the data was
    # serialized from. Special case: the pre-stable serialization format does not contain a version.
    version = cast(str, serialized["version"]) if "version" in serialized else None
    model: BaseModel = {
        None: PreStableModel,
        "1.0.0": BaseStateModel,
        "1.0.1": BaseStateModel
    }[version](**serialized)

    if isinstance(model, PreStableModel):
        # Run migrations from PreStableModel to StateModel
        bundle_needs_publish = bundle_needs_publish or model.changed

        model = BaseStateModel(
            identity_key=IdentityKeyPairModel(
                secret=base64.b64decode(model.ik.priv),
                secret_type=SecretType.PRIV
            ),
            signed_pre_key=SignedPreKeyPairModel(
                priv=base64.b64decode(model.spk.key.priv),
                sig=base64.b64decode(model.spk.signature),
                timestamp=int(model.spk.timestamp)
            ),
            old_signed_pre_key=None,
            pre_keys=frozenset({ base64.b64decode(pre_key.priv) for pre_key in model.otpks })
        )

    # Once all migrations have been applied, the model should be an instance of the most recent model
    assert isinstance(model, BaseStateModel)

    return model, bundle_needs_publish



import base64
import json
import os
import random
import time
from typing import Any, Dict, Iterator, List, Optional, Type, Union
from unittest import mock

import x3dh


__all__ = [
    "test_configuration",
    "test_key_agreements",
    "test_migrations",
    "test_old_signed_pre_key",
    "test_pre_key_availability",
    "test_pre_key_refill",
    "test_serialization",
    "test_signed_pre_key_rotation",
    "test_signed_pre_key_signature_verification"
]


try:
    import pytest
except ImportError:
    pass
else:
    pytestmark = pytest.mark.asyncio


def flip_random_bit(data: bytes, exclude_msb: bool = False) -> bytes:
    """
    Flip a random bit in a byte array.

    Args:
        data: The byte array to flip a random bit in.
        exclude_msb: Whether the most significant bit of the byte array should be excluded from the random
            selection. See note below.

    For Curve25519, the most significant bit of the public key is always cleared/ignored, as per RFC 7748 (on
    page 7). Thus, a bit flip of that bit does not make the signature verification fail, because the bit flip
    is ignored. The `exclude_msb` parameter can be used to disallow the bit flip to appear on the most
    significant bit and should be set when working with Curve25519 public keys.

    Returns:
        The data with a random bit flipped.
    """

    while True:
        modify_byte = random.randrange(len(data))
        modify_bit = random.randrange(8)

        # If the most significant bit was randomly chosen and `exclude_msb` is set, choose again.
        if not (exclude_msb and modify_byte == len(data) - 1 and modify_bit == 7):
            break

    data_mut = bytearray(data)
    data_mut[modify_byte] ^= 1 << modify_bit
    return bytes(data_mut)


bundles: Dict[bytes, x3dh.Bundle] = {}


class ExampleState(x3dh.State):
    """
    A state implementation for testing, which simulates bundle uploads by storing them in a global variable,
    and does some fancy public key encoding.
    """

    def _publish_bundle(self, bundle: x3dh.Bundle) -> None:
        bundles[bundle.identity_key] = bundle

    @staticmethod
    def _encode_public_key(key_format: x3dh.IdentityKeyFormat, pub: bytes) -> bytes:
        return b"\x42" + pub + b"\x13\x37" + key_format.value.encode("ASCII")


def get_bundle(state: ExampleState) -> x3dh.Bundle:
    """
    Retrieve a bundle from the simulated server.

    Args:
        state: The state to retrieve the bundle for.

    Returns:
        The bundle.

    Raises:
        AssertionError: if the bundle was never "uploaded".
    """

    if state.bundle.identity_key in bundles:
        return bundles[state.bundle.identity_key]
    assert False


def create_state(state_settings: Dict[str, Any]) -> ExampleState:
    """
    Create an :class:`ExampleState` and make sure the state creation worked as expected.

    Args:
        state_settings: Arguments to pass to :meth:`ExampleState.create`.

    Returns:
        The state.

    Raises:
        AssertionError: in case of failure.
    """

    exc: Optional[BaseException] = None
    state: Optional[ExampleState] = None
    try:
        state = ExampleState.create(**state_settings)
    except BaseException as e:  # pylint: disable=broad-except
        exc = e
    assert exc is None
    assert state is not None
    get_bundle(state)

    return state


def create_state_expect(
    state_settings: Dict[str, Any],
    expected_exception: Type[BaseException],
    expected_message: Union[str, List[str]]
) -> None:
    """
    Create an :class:`ExampleState`, but expect an exception to be raised during creation..

    Args:
        state_settings: Arguments to pass to :meth:`ExampleState.create`.
        expected_exception: The exception type expected to be raised.
        expected_message: The message expected to be raised, or a list of message snippets that should be part
            of the exception message.

    Raises:
        AssertionError: in case of failure.
    """

    exc: Optional[BaseException] = None
    state: Optional[ExampleState] = None
    try:
        state = ExampleState.create(**state_settings)
    except BaseException as e:  # pylint: disable=broad-except
        exc = e
    assert state is None

    assert isinstance(exc, expected_exception)
    if not isinstance(expected_message, list):
        expected_message = [ expected_message ]
    for expected_message_snippet in expected_message:
        assert expected_message_snippet in str(exc)


def generate_settings(
    info: bytes,
    signed_pre_key_rotation_period: int = 7 * 24 * 60 * 60,
    pre_key_refill_threshold: int = 25,
    pre_key_refill_target: int = 100
) -> Iterator[Dict[str, Any]]:
    """
    Generate state creation arguments.

    Args:
        info: The info to use constantly.
        signed_pre_key_rotation_period: The signed pre key rotation period to use constantly.
        pre_key_refill_threshold: The pre key refill threshold to use constantly.
        pre_key_refill_target. The pre key refill target to use constantly.

    Returns:
        An iterator which yields a set of state creation arguments, returning all valid combinations of
        identity key format and hash function with the given constant values.
    """

    for identity_key_format in [ x3dh.IdentityKeyFormat.CURVE_25519, x3dh.IdentityKeyFormat.ED_25519 ]:
        for hash_function in [ x3dh.HashFunction.SHA_256, x3dh.HashFunction.SHA_512 ]:
            state_settings: Dict[str, Any] = {
                "identity_key_format": identity_key_format,
                "hash_function": hash_function,
                "info": info,
                "signed_pre_key_rotation_period": signed_pre_key_rotation_period,
                "pre_key_refill_threshold": pre_key_refill_threshold,
                "pre_key_refill_target": pre_key_refill_target
            }

            yield state_settings


async def test_key_agreements() -> None:
    """
    Test the general key agreement functionality.
    """

    for state_settings in generate_settings("test_key_agreements".encode("ASCII")):
        state_a = create_state(state_settings)
        state_b = create_state(state_settings)

        # Store the current bundles
        bundle_a_before = get_bundle(state_a)
        bundle_b_before = get_bundle(state_b)

        # Perform the first, active half of the key agreement
        shared_secret_active, associated_data_active, header = await state_a.get_shared_secret_active(
            bundle_b_before,
            "ad appendix".encode("ASCII")
        )

        # Perform the second, passive half of the key agreement
        shared_secret_passive, associated_data_passive, _ = await state_b.get_shared_secret_passive(
            header,
            "ad appendix".encode("ASCII")
        )

        # Store the current bundles
        bundle_a_after = get_bundle(state_a)
        bundle_b_after = get_bundle(state_b)

        # The bundle of the active party should remain unmodified:
        assert bundle_a_after == bundle_a_before

        # The bundle of the passive party should have been modified and published again:
        assert bundle_b_after != bundle_b_before

        # To be exact, only one pre key should have been removed from the bundle:
        assert bundle_b_after.identity_key == bundle_b_before.identity_key
        assert bundle_b_after.signed_pre_key == bundle_b_before.signed_pre_key
        assert bundle_b_after.signed_pre_key_sig == bundle_b_before.signed_pre_key_sig
        assert len(bundle_b_after.pre_keys) == len(bundle_b_before.pre_keys) - 1
        assert all(pre_key in bundle_b_before.pre_keys for pre_key in bundle_b_after.pre_keys)

        # Both parties should have derived the same shared secret and built the same
        # associated data:
        assert shared_secret_active == shared_secret_passive
        assert associated_data_active == associated_data_passive

        # It should not be possible to accept the same header again:
        try:
            await state_b.get_shared_secret_passive(
                header,
                "ad appendix".encode("ASCII")
            )
            assert False
        except x3dh.KeyAgreementException as e:
            assert "pre key" in str(e)
            assert "not available" in str(e)

        # If the key agreement does not use a pre key, it should be possible to accept the header
        # multiple times:
        bundle_b = get_bundle(state_b)
        bundle_b = x3dh.Bundle(
            identity_key=bundle_b.identity_key,
            signed_pre_key=bundle_b.signed_pre_key,
            signed_pre_key_sig=bundle_b.signed_pre_key_sig,
            pre_keys=frozenset()
        )

        shared_secret_active, associated_data_active, header = await state_a.get_shared_secret_active(
            bundle_b,
            require_pre_key=False
        )

        shared_secret_passive, associated_data_passive, _ = await state_b.get_shared_secret_passive(
            header,
            require_pre_key=False
        )
        assert shared_secret_active == shared_secret_passive
        assert associated_data_active == associated_data_passive

        shared_secret_passive, associated_data_passive, _ = await state_b.get_shared_secret_passive(
            header,
            require_pre_key=False
        )
        assert shared_secret_active == shared_secret_passive
        assert associated_data_active == associated_data_passive


async def test_configuration() -> None:
    """
    Test whether incorrect argument values are rejected correctly.
    """

    for state_settings in generate_settings("test_configuration".encode("ASCII")):
        # Before destorying the settings, make sure that the state could be created like that:
        create_state(state_settings)

        state_settings["info"] = "test_configuration".encode("ASCII")

        # Pass an invalid timeout for the signed pre key
        state_settings["signed_pre_key_rotation_period"] = 0
        create_state_expect(state_settings, ValueError, "signed_pre_key_rotation_period")
        state_settings["signed_pre_key_rotation_period"] = -random.randrange(1, 2**64)
        create_state_expect(state_settings, ValueError, "signed_pre_key_rotation_period")
        state_settings["signed_pre_key_rotation_period"] = 1

        # Pass an invalid combination of pre_key_refill_threshold and pre_key_refill_target
        # pre_key_refill_threshold too small
        state_settings["pre_key_refill_threshold"] = 0
        create_state_expect(state_settings, ValueError, "pre_key_refill_threshold")
        state_settings["pre_key_refill_threshold"] = 25

        # pre_key_refill_target too small
        state_settings["pre_key_refill_target"] = 0
        create_state_expect(state_settings, ValueError, "pre_key_refill_target")
        state_settings["pre_key_refill_target"] = 100

        # pre_key_refill_threshold above pre_key_refill_target
        state_settings["pre_key_refill_threshold"] = 100
        state_settings["pre_key_refill_target"] = 25
        create_state_expect(state_settings, ValueError, [
            "pre_key_refill_threshold",
            "pre_key_refill_target"
        ])
        state_settings["pre_key_refill_threshold"] = 25
        state_settings["pre_key_refill_target"] = 100

        # pre_key_refill_threshold equals pre_key_refill_target (this should succeed)
        state_settings["pre_key_refill_threshold"] = 25
        state_settings["pre_key_refill_target"] = 25
        create_state(state_settings)
        state_settings["pre_key_refill_threshold"] = 25
        state_settings["pre_key_refill_target"] = 100


async def test_pre_key_refill() -> None:
    """
    Test pre key refill.
    """

    for state_settings in generate_settings(
        "test_pre_key_refill".encode("ASCII"),
        pre_key_refill_threshold=5,
        pre_key_refill_target=10
    ):
        state_a = create_state(state_settings)
        state_b = create_state(state_settings)

        # Verify that the bundle contains 100 pre keys initially:
        prev = len(get_bundle(state_b).pre_keys)
        assert prev == state_settings["pre_key_refill_target"]

        # Perform a lot of key agreements and verify that the refill works as expected:
        for _ in range(100):
            header = (await state_a.get_shared_secret_active(get_bundle(state_b)))[2]
            await state_b.get_shared_secret_passive(header)

            num_pre_keys = len(get_bundle(state_b).pre_keys)

            if prev == state_settings["pre_key_refill_threshold"]:
                assert num_pre_keys == state_settings["pre_key_refill_target"]
            else:
                assert num_pre_keys == prev - 1

            prev = num_pre_keys


async def test_signed_pre_key_signature_verification() -> None:
    """
    Test signature verification of the signed pre key.
    """

    for state_settings in generate_settings("test_signed_pre_key_signature_verification".encode("ASCII")):
        identity_key_format: x3dh.IdentityKeyFormat = state_settings["identity_key_format"]

        for _ in range(8):
            state_a = create_state(state_settings)
            state_b = create_state(state_settings)

            bundle = get_bundle(state_b)

            # First, make sure that the active half of the key agreement usually works:
            await state_a.get_shared_secret_active(bundle)

            # Now, flip a random bit in
            # 1. the signature
            # 2. the identity key
            # 3. the signed pre key
            # and make sure that the active half of the key agreement reject the signature.

            # 1.: the signature
            signed_pre_key_sig = flip_random_bit(bundle.signed_pre_key_sig)
            bundle_modified = x3dh.Bundle(
                identity_key=bundle.identity_key,
                signed_pre_key=bundle.signed_pre_key,
                signed_pre_key_sig=signed_pre_key_sig,
                pre_keys=bundle.pre_keys
            )
            try:
                await state_a.get_shared_secret_active(bundle_modified)
                assert False
            except x3dh.KeyAgreementException as e:
                assert "signature" in str(e)

            # 2.: the identity key
            exclude_msb = identity_key_format is x3dh.IdentityKeyFormat.CURVE_25519
            identity_key = flip_random_bit(bundle.identity_key, exclude_msb=exclude_msb)
            bundle_modified = x3dh.Bundle(
                identity_key=identity_key,
                signed_pre_key=bundle.signed_pre_key,
                signed_pre_key_sig=bundle.signed_pre_key_sig,
                pre_keys=bundle.pre_keys
            )
            try:
                await state_a.get_shared_secret_active(bundle_modified)
                assert False
            except x3dh.KeyAgreementException as e:
                assert "signature" in str(e)

            # 3.: the signed pre key
            signed_pre_key = flip_random_bit(bundle.signed_pre_key)
            bundle_modified = x3dh.Bundle(
                identity_key=bundle.identity_key,
                signed_pre_key=signed_pre_key,
                signed_pre_key_sig=bundle.signed_pre_key_sig,
                pre_keys=bundle.pre_keys
            )
            try:
                await state_a.get_shared_secret_active(bundle_modified)
                assert False
            except x3dh.KeyAgreementException as e:
                assert "signature" in str(e)


async def test_pre_key_availability() -> None:
    """
    Test whether key agreements without pre keys work/are rejected as expected.
    """

    for state_settings in generate_settings("test_pre_key_availability".encode("ASCII")):
        state_a = create_state(state_settings)
        state_b = create_state(state_settings)

        # First, test the active half of the key agreement
        for require_pre_key in [ True, False ]:
            for include_pre_key in [ True, False ]:
                bundle = get_bundle(state_b)

                # Make sure that the bundle contains pre keys:
                assert len(bundle.pre_keys) > 0

                # If required for the test, remove all pre keys:
                if not include_pre_key:
                    bundle = x3dh.Bundle(
                        identity_key=bundle.identity_key,
                        signed_pre_key=bundle.signed_pre_key,
                        signed_pre_key_sig=bundle.signed_pre_key_sig,
                        pre_keys=frozenset()
                    )

                should_fail = require_pre_key and not include_pre_key
                try:
                    header = (await state_a.get_shared_secret_active(
                        bundle,
                        require_pre_key=require_pre_key
                    ))[2]
                    assert not should_fail
                    assert (header.pre_key is not None) == include_pre_key
                except x3dh.KeyAgreementException as e:
                    assert should_fail
                    assert "does not contain" in str(e)
                    assert "pre key" in str(e)

        # Second, test the passive half of the key agreement
        for require_pre_key in [ True, False ]:
            for include_pre_key in [ True, False ]:
                bundle = get_bundle(state_b)

                # Make sure that the bundle contains pre keys:
                assert len(bundle.pre_keys) > 0

                # If required for the test, remove all pre keys:
                if not include_pre_key:
                    bundle = x3dh.Bundle(
                        identity_key=bundle.identity_key,
                        signed_pre_key=bundle.signed_pre_key,
                        signed_pre_key_sig=bundle.signed_pre_key_sig,
                        pre_keys=frozenset()
                    )

                # Perform the active half of the key agreement, using a pre key only if required for
                # the test.
                shared_secret_active, _, header = await state_a.get_shared_secret_active(
                    bundle,
                    require_pre_key=False
                )

                should_fail = require_pre_key and not include_pre_key
                try:
                    shared_secret_passive, _, _ = await state_b.get_shared_secret_passive(
                        header,
                        require_pre_key=require_pre_key
                    )
                    assert not should_fail
                    assert shared_secret_passive == shared_secret_active
                except x3dh.KeyAgreementException as e:
                    assert should_fail
                    assert "does not use" in str(e)
                    assert "pre key" in str(e)


THREE_DAYS = 3 * 24 * 60 * 60
EIGHT_DAYS = 8 * 24 * 60 * 60


async def test_signed_pre_key_rotation() -> None:
    """
    Test signed pre key rotation logic.
    """

    for state_settings in generate_settings("test_signed_pre_key_rotation".encode("ASCII")):
        state_b = create_state(state_settings)
        bundle_b = get_bundle(state_b)

        current_time = time.time()
        time_mock = mock.MagicMock()

        # Mock time.time, so that the test can skip days in an instant
        with mock.patch("time.time", time_mock):
            # ExampleState.create should call time.time only once, when generating the signed pre key. Make
            # the mock return the actual current time for that call.
            time_mock.return_value = current_time
            state_a = create_state(state_settings)
            assert time_mock.call_count == 1
            time_mock.reset_mock()

            # Prepare a key agreement header, the time is irrelevant here. Don't use a pre key so
            # that the header can be used multiple times.
            bundle_a = get_bundle(state_a)
            bundle_a = x3dh.Bundle(
                identity_key=bundle_a.identity_key,
                signed_pre_key=bundle_a.signed_pre_key,
                signed_pre_key_sig=bundle_a.signed_pre_key_sig,
                pre_keys=frozenset()
            )

            time_mock.return_value = current_time + THREE_DAYS
            header_b = (await state_b.get_shared_secret_active(bundle_a, require_pre_key=False))[2]
            state_b.rotate_signed_pre_key()
            assert time_mock.call_count == 1
            time_mock.reset_mock()

            # There are three methods that check whether the signed pre key has to be rotated:
            # 1. get_shared_secret_active
            # 2. get_shared_secret_passive
            # 3. deserialize

            # 1. get_shared_secret_active

            # Make the mock return the actual current time plus three days. This should not trigger a
            # rotation.
            bundle_a_before = get_bundle(state_a)
            time_mock.return_value = current_time + THREE_DAYS
            await state_a.get_shared_secret_active(bundle_b)
            state_a.rotate_signed_pre_key()
            assert time_mock.call_count == 1
            time_mock.reset_mock()
            assert get_bundle(state_a) == bundle_a_before

            # Make the mock return the actual current time plus eight days. This should trigger a rotation.
            # A rotation reads the time twice.
            bundle_a_before = get_bundle(state_a)
            time_mock.return_value = current_time + EIGHT_DAYS
            await state_a.get_shared_secret_active(bundle_b)
            state_a.rotate_signed_pre_key()
            assert time_mock.call_count == 2
            time_mock.reset_mock()
            assert get_bundle(state_a).identity_key == bundle_a_before.identity_key
            assert get_bundle(state_a).signed_pre_key != bundle_a_before.signed_pre_key
            assert get_bundle(state_a).signed_pre_key_sig != bundle_a_before.signed_pre_key_sig
            assert get_bundle(state_a).pre_keys == bundle_a_before.pre_keys

            # Update the "current_time" to the creation time of the last signed pre key:
            current_time += EIGHT_DAYS

            # 2. get_shared_secret_passive

            # Make the mock return the actual current time plus three days. This should not trigger a
            # rotation.
            bundle_a_before = get_bundle(state_a)
            time_mock.return_value = current_time + THREE_DAYS
            await state_a.get_shared_secret_passive(header_b, require_pre_key=False)
            state_a.rotate_signed_pre_key()
            assert time_mock.call_count == 1
            time_mock.reset_mock()
            assert get_bundle(state_a) == bundle_a_before

            # Make the mock return the actual current time plus eight days. This should trigger a rotation.
            # A rotation reads the time twice.
            bundle_a_before = get_bundle(state_a)
            time_mock.return_value = current_time + EIGHT_DAYS
            await state_a.get_shared_secret_passive(header_b, require_pre_key=False)
            state_a.rotate_signed_pre_key()
            assert time_mock.call_count == 2
            time_mock.reset_mock()
            assert get_bundle(state_a).identity_key == bundle_a_before.identity_key
            assert get_bundle(state_a).signed_pre_key != bundle_a_before.signed_pre_key
            assert get_bundle(state_a).signed_pre_key_sig != bundle_a_before.signed_pre_key_sig
            assert get_bundle(state_a).pre_keys == bundle_a_before.pre_keys

            # Update the "current_time" to the creation time of the last signed pre key:
            current_time += EIGHT_DAYS

            # 3. deserialize

            # Make the mock return the actual current time plus three days. This should not trigger a
            # rotation.
            bundle_a_before = get_bundle(state_a)
            time_mock.return_value = current_time + THREE_DAYS
            state_a = ExampleState.from_model(state_a.model, **state_settings)
            assert time_mock.call_count == 1
            time_mock.reset_mock()
            assert get_bundle(state_a) == bundle_a_before

            # Make the mock return the actual current time plus eight days. This should trigger a rotation.
            # A rotation reads the time twice.
            bundle_a_before = get_bundle(state_a)
            time_mock.return_value = current_time + EIGHT_DAYS
            state_a = ExampleState.from_model(state_a.model, **state_settings)
            assert time_mock.call_count == 2
            time_mock.reset_mock()
            assert get_bundle(state_a).identity_key == bundle_a_before.identity_key
            assert get_bundle(state_a).signed_pre_key != bundle_a_before.signed_pre_key
            assert get_bundle(state_a).signed_pre_key_sig != bundle_a_before.signed_pre_key_sig
            assert get_bundle(state_a).pre_keys == bundle_a_before.pre_keys

            # Update the "current_time" to the creation time of the last signed pre key:
            current_time += EIGHT_DAYS


async def test_old_signed_pre_key() -> None:
    """
    Test that the old signed pre key remains available for key agreements for one further rotation period.
    """

    for state_settings in generate_settings(
        "test_old_signed_pre_key".encode("ASCII"),
        signed_pre_key_rotation_period=2
    ):
        print(state_settings)
        state_a = create_state(state_settings)
        state_b = create_state(state_settings)

        # Prepare a key agreement header using the current signed pre key of state a. Don't use a pre
        # key so that the header can be used multiple times.
        bundle_a = get_bundle(state_a)
        bundle_a_no_pre_keys = x3dh.Bundle(
            identity_key=bundle_a.identity_key,
            signed_pre_key=bundle_a.signed_pre_key,
            signed_pre_key_sig=bundle_a.signed_pre_key_sig,
            pre_keys=frozenset()
        )
        shared_secret_active, associated_data_active, header = await state_b.get_shared_secret_active(
            bundle_a_no_pre_keys,
            require_pre_key=False
        )

        # Make sure that this key agreement works as intended:
        shared_secret_passive, associated_data_passive, _ = await state_a.get_shared_secret_passive(
            header,
            require_pre_key=False
        )
        assert shared_secret_active == shared_secret_passive
        assert associated_data_active == associated_data_passive

        # Rotate the signed pre key once. The rotation period is specified as two days, still skipping eight
        # days should only trigger a single rotation.
        current_time = time.time()
        time_mock = mock.MagicMock()

        # Mock time.time, so that the test can skip days in an instant
        with mock.patch("time.time", time_mock):
            time_mock.return_value = current_time + EIGHT_DAYS
            state_a = ExampleState.from_model(state_a.model, **state_settings)
            assert time_mock.call_count == 2
            time_mock.reset_mock()

        # Make sure that the signed pre key was rotated:
        assert get_bundle(state_a).identity_key == bundle_a.identity_key
        assert get_bundle(state_a).signed_pre_key != bundle_a.signed_pre_key
        assert get_bundle(state_a).signed_pre_key_sig != bundle_a.signed_pre_key_sig
        assert get_bundle(state_a).pre_keys == bundle_a.pre_keys

        bundle_a_rotated = get_bundle(state_a)

        # The old signed pre key should still be stored in state_a, thus the old key agreement header should
        # still work:
        shared_secret_passive, associated_data_passive, _ = await state_a.get_shared_secret_passive(
            header,
            require_pre_key=False
        )
        assert shared_secret_active == shared_secret_passive
        assert associated_data_active == associated_data_passive

        # Rotate the signed pre key again:
        with mock.patch("time.time", time_mock):
            time_mock.return_value = current_time + EIGHT_DAYS + THREE_DAYS
            state_a = ExampleState.from_model(state_a.model, **state_settings)
            assert time_mock.call_count == 2
            time_mock.reset_mock()

        # Make sure that the signed pre key was rotated again:
        assert get_bundle(state_a).identity_key == bundle_a.identity_key
        assert get_bundle(state_a).signed_pre_key != bundle_a.signed_pre_key
        assert get_bundle(state_a).signed_pre_key_sig != bundle_a.signed_pre_key_sig
        assert get_bundle(state_a).pre_keys == bundle_a.pre_keys
        assert get_bundle(state_a).identity_key == bundle_a_rotated.identity_key
        assert get_bundle(state_a).signed_pre_key != bundle_a_rotated.signed_pre_key
        assert get_bundle(state_a).signed_pre_key_sig != bundle_a_rotated.signed_pre_key_sig
        assert get_bundle(state_a).pre_keys == bundle_a_rotated.pre_keys

        # Now the signed pre key used in the header should not be available any more, the passive half of the
        # key agreement should fail:
        try:
            await state_a.get_shared_secret_passive(header, require_pre_key=False)
            assert False
        except x3dh.KeyAgreementException as e:
            assert "signed pre key" in str(e)
            assert "not available" in str(e)


async def test_serialization() -> None:
    """
    Test (de)serialization.
    """

    for state_settings in generate_settings("test_serialization".encode("ASCII")):
        state_a = create_state(state_settings)
        state_b = create_state(state_settings)

        # Make sure that the key agreement works normally:
        shared_secret_active, associated_data_acitve, header = await state_a.get_shared_secret_active(
            get_bundle(state_b)
        )
        shared_secret_passive, associated_data_passive, _ = await state_b.get_shared_secret_passive(header)
        assert shared_secret_active == shared_secret_passive
        assert associated_data_acitve == associated_data_passive

        # Do the same thing but serialize and deserialize state b before performing the passive half of the
        # key agreement:
        bundle_b_before = get_bundle(state_b)

        shared_secret_active, associated_data_acitve, header = await state_a.get_shared_secret_active(
            get_bundle(state_b)
        )
        state_b = ExampleState.from_model(state_b.model, **state_settings)
        shared_secret_passive, associated_data_passive, _ = await state_b.get_shared_secret_passive(header)
        assert shared_secret_active == shared_secret_passive
        assert associated_data_acitve == associated_data_passive

        # Make sure that the bundle remained the same, except for one pre key being deleted:
        assert get_bundle(state_b).identity_key == bundle_b_before.identity_key
        assert get_bundle(state_b).signed_pre_key == bundle_b_before.signed_pre_key
        assert get_bundle(state_b).signed_pre_key_sig == bundle_b_before.signed_pre_key_sig
        assert len(get_bundle(state_b).pre_keys) == len(bundle_b_before.pre_keys) - 1
        assert all(pre_key in bundle_b_before.pre_keys for pre_key in get_bundle(state_b).pre_keys)

        # Accepting a key agreement using a pre key results in the pre key being deleted
        # from the state. Use (de)serialization to circumvent the deletion of the pre key. This time
        # also serialize the structure into JSON:
        shared_secret_active, associated_data_acitve, header = await state_a.get_shared_secret_active(
            get_bundle(state_b)
        )
        state_b_serialized = json.dumps(state_b.json)

        # Accepting the header should work once...
        shared_secret_passive, associated_data_passive, _ = await state_b.get_shared_secret_passive(header)
        assert shared_secret_active == shared_secret_passive
        assert associated_data_acitve == associated_data_passive

        # ...but fail the second time:
        try:
            await state_b.get_shared_secret_passive(header)
            assert False
        except x3dh.KeyAgreementException as e:
            assert "pre key" in str(e)
            assert "not available" in str(e)

        # After restoring the state, it should work again:
        state_b, needs_publish = ExampleState.from_json(json.loads(state_b_serialized), **state_settings)
        shared_secret_passive, associated_data_passive, _ = await state_b.get_shared_secret_passive(header)
        assert not needs_publish
        assert shared_secret_active == shared_secret_passive
        assert associated_data_acitve == associated_data_passive


THIS_FILE_PATH = os.path.dirname(os.path.abspath(__file__))


async def test_migrations() -> None:
    """
    Test the migration from pre-stable.
    """

    state_settings: Dict[str, Any] = {
        "identity_key_format": x3dh.IdentityKeyFormat.CURVE_25519,
        "hash_function": x3dh.HashFunction.SHA_256,
        "info": "test_migrations".encode("ASCII"),
        "signed_pre_key_rotation_period": 7,
        "pre_key_refill_threshold": 25,
        "pre_key_refill_target": 100
    }

    with open(os.path.join(
        THIS_FILE_PATH,
        "migration_data",
        "state-alice-pre-stable.json"
    ), "r", encoding="utf-8") as state_alice_pre_stable_json:
        state_a_serialized = json.load(state_alice_pre_stable_json)

    with open(os.path.join(
        THIS_FILE_PATH,
        "migration_data",
        "state-bob-pre-stable.json"
    ), "r", encoding="utf-8") as state_bob_pre_stable_json:
        state_b_serialized = json.load(state_bob_pre_stable_json)

    with open(os.path.join(
        THIS_FILE_PATH,
        "migration_data",
        "shared-secret-pre-stable.json"
    ), "r", encoding="utf-8") as shared_secret_pey_stable_json:
        shared_secret_active_serialized = json.load(shared_secret_pey_stable_json)

    # Convert the pre-stable shared secret structure into a x3dh.SharedSecretActive
    shared_secret_active = base64.b64decode(shared_secret_active_serialized["sk"].encode("ASCII"))
    associated_data_active = base64.b64decode(shared_secret_active_serialized["ad"].encode("ASCII"))
    header = x3dh.Header(
        identity_key=base64.b64decode(shared_secret_active_serialized["to_other"]["ik"].encode("ASCII")),
        ephemeral_key=base64.b64decode(shared_secret_active_serialized["to_other"]["ek"].encode("ASCII")),
        signed_pre_key=base64.b64decode(shared_secret_active_serialized["to_other"]["spk"].encode("ASCII")),
        pre_key=base64.b64decode(shared_secret_active_serialized["to_other"]["otpk"].encode("ASCII"))
    )

    # Load state a. This should not trigger a publishing of the bundle, as the `changed` flag is not set.
    state_a, _needs_publish = ExampleState.from_json(state_a_serialized, **state_settings)

    try:
        get_bundle(state_a)
        assert False
    except AssertionError:
        pass

    # Load state b. This should trigger a publishing of the bundle, as the `changed` flag is set.
    state_b, _needs_publish = ExampleState.from_json(state_b_serialized, **state_settings)

    get_bundle(state_b)

    # Complete the passive half of the key agreement as created by the pre-stable version:
    shared_secret_passive, associated_data_passive, _ = await state_b.get_shared_secret_passive(header)
    assert shared_secret_active == shared_secret_passive
    # Don't check the associated data, since formats have changed.

    # Try another key agreement using the migrated sessions:
    shared_secret_active, associated_data_active, header = await state_a.get_shared_secret_active(
        get_bundle(state_b)
    )
    shared_secret_passive, associated_data_passive, _ = await state_b.get_shared_secret_passive(header)
    assert shared_secret_active == shared_secret_passive
    assert associated_data_active == associated_data_passive



from __future__ import annotations

import enum
from typing import FrozenSet, List, Mapping, NamedTuple, Optional, Union
from typing_extensions import TypeAlias


__all__ = [
    "Bundle",
    "IdentityKeyFormat",
    "Header",
    "JSONType",
    "JSONObject",
    "SecretType"
]


################
# Type Aliases #
################

JSONType: TypeAlias = Union[Mapping[str, "JSONType"], List["JSONType"], str, int, float, bool, None]
JSONObject: TypeAlias = Mapping[str, "JSONType"]


############################
# Structures (NamedTuples) #
############################

class Bundle(NamedTuple):
    """
    The bundle is a collection of public keys and signatures used by the X3DH protocol to achieve asynchronous
    key agreements while providing forward secrecy and cryptographic deniability. Parties that want to be
    available for X3DH key agreements have to publish their bundle somehow. Other parties can then use that
    bundle to perform a key agreement.
    """

    identity_key: bytes
    signed_pre_key: bytes
    signed_pre_key_sig: bytes
    pre_keys: FrozenSet[bytes]


class Header(NamedTuple):
    """
    The header generated by the active party as part of the key agreement, and consumed by the passive party
    to derive the same shared secret.
    """

    identity_key: bytes
    ephemeral_key: bytes
    signed_pre_key: bytes
    pre_key: Optional[bytes]


################
# Enumerations #
################

@enum.unique
class IdentityKeyFormat(enum.Enum):
    """
    The two supported public key formats for the identity key:

    * Curve25519 public keys: 32 bytes, the little-endian encoding of the u coordinate as per `RFC 7748,
      section 5 "The X25519 and X448 Functions" <https://www.rfc-editor.org/rfc/rfc7748.html#section-5>`_.
    * Ed25519 public keys: 32 bytes, the little-endian encoding of the y coordinate with the sign bit of the x
      coordinate stored in the most significant bit as per `RFC 8032, section 3.2 "Keys"
      <https://www.rfc-editor.org/rfc/rfc8032.html#section-3.2>`_.
    """

    CURVE_25519 = "CURVE_25519"
    ED_25519 = "ED_25519"


@enum.unique
class SecretType(enum.Enum):
    """
    The two types of secrets that an :class:`~x3dh.identity_key_pair.IdentityKeyPair` can use internally: a
    seed or a private key.
    """

    SEED = "SEED"
    PRIV = "PRIV"
